/** \file   htcbdd.c
 *  \brief  Hypergraph Transversal Computation with BDDs and ZDDs
 *  \author Takahisa Toda
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "my_def.h"
#include "my_random.h"
#include "my_cache.h"
#include "dd_io_setfam.h"
#include "zdd_extra.h"
#include "dd_hit.h"
#include "dd_min.h"

#define PROGNAME "HTCBDD"//!< the name of this program
#define PROGVERS "1.1.0" //!< the current version

#ifdef SELECT_CUDD
DdManager *dd_mgr = NULL; //!< BDD/ZDD manager for CUDD
#endif /*SELECT_CUDD*/

#ifndef NGSL
my_rng *g_rng = NULL; //!< random number generator for GSL
#endif /*NGSL*/

#define PRINT_USAGE(p) do{fprintf(stderr, "Usage: %s [option] input-file [output-file]\n", (p)); \
    fprintf(stderr, "-t Toda method (default)\n");			\
    fprintf(stderr, "-k Knuth method\n");				\
  } while(0)

#define T_METHOD 't'
#define K_METHOD 'k'

int main(int argc, char *argv[])
{   
  int res = MY_SUCCESS;
  char method = T_METHOD;
  char *infile = NULL;
  char *outfile = NULL;

  /*** INPUT ARGUMENT ***/  
  for(int i = 1; i < argc; i++) {
    if(argv[i][0] == '-') {
      switch (argv[i][1]){
      case T_METHOD: case K_METHOD:
	method = argv[i][1];
	break; 
      case 'v':
	fprintf(stderr, "%s\tversion %s\n\n", PROGNAME, PROGVERS);
	return MY_SUCCESS;
      case '?': case 'h': default:
	PRINT_USAGE(argv[0]); return  MY_SUCCESS;
      }
    } else {
      if(infile == NULL) {
	infile = argv[i];
	if(infile == NULL) {
	  fprintf(stderr, "error: the path(%s) does not exist\n", argv[i]);
	  exit(EXIT_FAILURE);
	}
      }
      else if(outfile == NULL) {outfile = argv[i];}
      else {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
    }
  }
  if(infile == NULL) {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("%s\t\tversion %s\n", PROGNAME, PROGVERS);
  printf("package\t\t%s\n", BDD_PACKAGE);
  printf("package\t\t%s\n", ZDD_PACKAGE);
  printf("input\t\t%s\n", infile);
  if(outfile != NULL) printf("output\t\t%s\n", outfile);
  if(method == T_METHOD) {
    printf("method\t\tToda\n");
  } else if(method == K_METHOD) {
    printf("method\t\tKnuth\n");
  } else {
    exit(EXIT_FAILURE);
  }
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/

  my_initrand(); // initialize random number generator
  FILE *in = fopen(infile, "r");
  ENSURE_TRUE_MSG(in != NULL, "file open failed\n");

  /*** INITIALIZE BDD/ZDD MANAGERS ***/
  size_t nrow, maxval, total_nent;
  res = getfileinfo(&nrow, &maxval, &total_nent, in);
  ENSURE_SUCCESS(res);
  // adjust CACHE_DEFAULT_SIZE if necessary.
  bdd_init(maxval, CACHE_DEFAULT_SIZE);
  zdd_init(maxval, CACHE_DEFAULT_SIZE);
  zdd_init_extra(CACHE_DEFAULT_SIZE);

  /*** IMPORT ZDD ***/  
  clock_t startingtime = my_clock();
  my_zddp f = zimport(nrow, total_nent, in);
  ENSURE_TRUE_MSG(f != ZDD_NULL, "ZDD import failed\n");
  print_elapsed_time(startingtime, "IMPORT");
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("|zdd|\t\t%zu\n", zdd_size(f));
  printf("#sets\t\t%zu\n\n", zdd_card(f));
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
  
  /*** TRANSVERSAL HYPERGRAPH COMPUTATION ***/
  my_zddp g = ZDD_NULL;
  switch(method) {
  case T_METHOD:
    startingtime = my_clock();
    my_bddp t = hit_z2b(f);
    ENSURE_TRUE(t != BDD_NULL);
    print_elapsed_time(startingtime, "HIT");
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
    printf("|bdd|\t\t%zu\n\n", bdd_size(t));
    fflush(stdout);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
    
    startingtime = my_clock();
    g = minimal_b2z_monotone(t);
    ENSURE_TRUE(g != ZDD_NULL);
    print_elapsed_time(startingtime, "MIN");
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
    printf("|zdd|\t\t%zu\n", zdd_size(g));
    printf("#sets\t\t%zu\n\n", zdd_card(g));
    fflush(stdout);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
    break;
    
  case K_METHOD:
    startingtime = my_clock();
    g = minhit_nonsup(f);
    ENSURE_TRUE(g != ZDD_NULL);
    print_elapsed_time(startingtime, "MINHIT");
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
    printf("|zdd|\t\t%zu\n", zdd_size(g));
    printf("#sets\t\t%zu\n\n", zdd_card(g));
    fflush(stdout);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
    break;

  default:
    fprintf(stderr, "error: invalid computation mode\n");
    exit(EXIT_FAILURE);
    break;
  }

  /*** EXPORT ZDD ***/  
  if(outfile != NULL) {
    FILE *out = fopen(outfile, "w");
    startingtime = my_clock();
    zexport(maxval, g, out);
    print_elapsed_time(startingtime, "EXPORT");
    fclose(out);
  }

  bdd_quit(); zdd_quit(); zdd_quit_extra();
  my_freerand();
  fclose(in);

  return MY_SUCCESS;
}


