/** \file   my_hash.c
 *  \brief  Implementation of a simple hash table
 *  \author Takahisa Toda
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "my_def.h"
#include "my_hash.h"


/** \brief  Create a hash table.
 *  \param  nbuckets number of buckets
 *  \return pointer to a hash table if successful; otherwise, NULL.
 *  \see    ht_destroy, ht_enlarge
 */
struct htable *ht_create(size_t nbuckets)
{
  ENSURE_TRUE(nbuckets != 0);
  struct htable *h = (struct htable*) malloc(sizeof(struct htable));
  ENSURE_TRUE_MSG(h != NULL,"memory allocation failed");

  h->nbucket = nbuckets;
  h->bucket = (struct htable_entry**) malloc(sizeof(struct htable_entry*) * nbuckets);
  ENSURE_TRUE_MSG(h->bucket != NULL, "memory allocation failed");
  for (size_t i = 0; i < nbuckets; i++) h->bucket[i] = NULL;

  return h;
}


/** \brief Free memory area used by a hash table.
 *  \param h pointer to a hash table
 *  \see   ht_create, ht_enlarge
 */
void ht_destroy(struct htable *h)
{
  assert(h != NULL);
  const size_t nbucket = h->nbucket;
  for(size_t i = 0; i < nbucket; i++) {
      for(struct htable_entry *c = h->bucket[i]; c != NULL; ) {
	struct htable_entry *t = c->next;
	free(c);
	c = t;
      }
  }
  free(h->bucket); h->bucket = NULL;
  free(h); h = NULL;
}

/** \brief  Calculate a hash value.
 *  \param  k  key
 *  \param  size the size of a hash table, i.e. the number of buckets
 *  \return the calculated hash value
 */
static inline size_t my_hash(size_t k, size_t size)
{
  assert(size != 0);
  return (k*MY_PRIME(0))%size;
}


/** \brief  Register a new entry.
 *  \param  k  key
 *  \param  d  data associated with key
 *  \param  h  pointer to a hash table
 *  \see    ht_search
 */
void ht_add(size_t k, size_t d, struct htable *h)
{
  assert(h != NULL);
  struct htable_entry *c = (struct htable_entry*)malloc(sizeof(struct htable_entry));
  ENSURE_TRUE_MSG(c != NULL, "memory allocation failed");
  c->key = k;
  c->data = d;

  const size_t val = my_hash(k, h->nbucket);
  c->next = h->bucket[val];
  h->bucket[val] = c;
}


/** \brief  Find a hash entry. If found, copy the associated data to pd.
 *  \param  k    key
 *  \param  pd   pointer to a variable, in which data is copied.
 *  \param  h    pointer to a hash table
 *  \return MY_SUCCESS if found; otherwise, MY_FAILURE.
 *  \see    ht_add
 *  \note
 *  - data is not copied if pd == NULL.
 */
int ht_search(size_t k, size_t *pd, const struct htable *h)
{
  assert(h != NULL);
  const size_t val = my_hash(k, h->nbucket);
  for(struct htable_entry *c = h->bucket[val]; c != NULL; c = c->next) {
    if(c->key == k) {
      if(pd != NULL) *pd = c->data;
      return MY_SUCCESS;
    }
  }
  return MY_FAILURE;
}


/* \brief Compute the maximum length of a linked list.
 *  \param h pointer to a hash table
 *  \return the computed maximum length
 *  \see ht_create, ht_destroy
 */
size_t ht_maxdepth(const struct htable *h)
{
  assert(h != NULL);
  size_t depth = 0;
  const size_t nbucket = h->nbucket;
  for(size_t i = 0; i < nbucket; i++) {
      size_t j = 0;
      for(struct htable_entry *t = h->bucket[i]; t != NULL; t=t->next) {j++;}
      if(depth < j) {depth = j;}
    }
  return depth;
}

/* \brief Enlarge hash table size
 * \param ch pointer to a hash table
 * \see ht_create, ht_destroy
 */
void ht_enlarge(struct htable *h)
{
  const size_t oldsiz = h->nbucket;
  const size_t newsiz = oldsiz << 2;
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("hash table resizing:%zu -> %zu\n", oldsiz, newsiz);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
  struct htable_entry **oldbucket = h->bucket;
  struct htable_entry **newbucket = (struct htable_entry**) malloc(sizeof(struct htable_entry*) * newsiz);
  ENSURE_TRUE_MSG(newbucket != NULL, "memory allocation failed");
  for(size_t i = 0; i < newsiz; i++) newbucket[i] = NULL;
  for(size_t i = 0; i < oldsiz; i++) {
    for(struct htable_entry *c = oldbucket[i]; c != NULL;) {
      struct htable_entry *nextnode = c->next;
      const size_t val = my_hash(c->key, newsiz);
      c->next = newbucket[val];
      newbucket[val] = c;      
      c = nextnode;
    }
  }
  free(oldbucket); oldbucket = NULL;
  h->nbucket = newsiz;
  h->bucket = newbucket;
}
